from __future__ import annotations

from pathlib import Path

import pandas as pd

from bionty.base._settings import settings
from bionty.base.dev._io import load_yaml


def LAMINDB_INSTANCE_LOADED():
    is_loaded = False
    lnenv_filepath = Path.home() / ".lamin/current_instance.env"
    if lnenv_filepath.exists():
        with open(lnenv_filepath.as_posix()) as f:
            is_loaded = "bionty" in f.read().split("schema_str=")[-1]
    return is_loaded


def parse_sources_yaml(
    filepath: str | Path = settings.public_sources,
) -> pd.DataFrame:
    """Parse values from sources yaml file into a DataFrame.

    Args:
        filepath: Path to the versions yaml file.

    Returns:
        - entity
        - name
        - organism
        - version
        - url
        - description
        - source_website
    """
    all_rows = []
    for entity, sources in load_yaml(filepath).items():
        if entity == "version":
            continue
        for source, organism_source in sources.items():
            name = organism_source.get("name", "")
            website = organism_source.get("website", "")
            for organism, versions in organism_source.items():
                if organism in ["name", "website"]:
                    continue
                for version_key, version_meta in versions.items():
                    row = (
                        entity,
                        source,
                        organism,
                        str(version_key),
                        version_meta.get("url"),
                        name,
                        website,
                    )
                    all_rows.append(row)

    return pd.DataFrame(
        all_rows,
        columns=[
            "entity",
            "name",
            "organism",
            "version",
            "url",
            "description",
            "source_website",
        ],
    )


def parse_currently_used_sources(yaml: str | Path | list[dict]) -> dict:
    """Parse out the most recent versions from yaml."""
    if isinstance(yaml, str | Path):
        df = parse_sources_yaml(yaml)
        df_current = (
            df[["entity", "name", "organism", "version"]]  # type: ignore
            .drop_duplicates(["entity", "organism", "name"], keep="first")
            .groupby(["entity", "organism", "name"], sort=False)
            .max()
        )
        records = df_current.reset_index().to_dict(orient="records")
    else:
        records = yaml

    current_dict: dict = {}
    for kwargs in records:
        entity, organism, source, version = (
            kwargs["entity"],
            kwargs["organism"],
            kwargs["name"],
            kwargs["version"],
        )
        if entity not in current_dict:
            current_dict[entity] = {}
        if organism not in current_dict[entity]:
            current_dict[entity][organism] = {source: version}
    return current_dict



import shutil

import lamindb_setup as ln_setup
import pytest


def pytest_sessionstart():
    ln_setup.init(storage="./testdb", modules="bionty,wetlab")


def pytest_sessionfinish(session: pytest.Session):
    shutil.rmtree("./testdb")
    ln_setup.delete("testdb", force=True)



from functools import wraps
from pathlib import Path
from typing import Union

HOME_DIR = Path(f"{Path.home()}/.lamin/bionty").resolve()
ROOT_DIR = Path(__file__).parent.resolve()


def check_datasetdir_exists(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        settings.datasetdir.mkdir(exist_ok=True)
        return f(*args, **kwargs)

    return wrapper


def check_dynamicdir_exists(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        settings.dynamicdir.mkdir(exist_ok=True)
        return f(*args, **kwargs)

    return wrapper


class Settings:
    def __init__(
        self,
        datasetdir: str | Path = ROOT_DIR / "data/",
        dynamicdir: str | Path = ROOT_DIR / "_dynamic/",
    ):
        # setters convert to Path and resolve:
        self.datasetdir = datasetdir
        self.dynamicdir = dynamicdir

    @property
    def datasetdir(self):
        """Directory for datasets."""
        return self._datasetdir

    @datasetdir.setter
    def datasetdir(self, datasetdir: str | Path):
        self._datasetdir = Path(datasetdir).resolve()

    @property
    def dynamicdir(self):
        """Directory for datasets."""
        return self._dynamicdir

    @dynamicdir.setter
    def dynamicdir(self, dynamicdir: str | Path):
        self._dynamicdir = Path(dynamicdir).resolve()

    @property
    def public_sources(self):
        return ROOT_DIR / "sources.yaml"


settings = Settings()



import bionty.base as bt_base
import pandas as pd


def test_ensembl_organism_inspect_name():
    df = pd.DataFrame(
        index=[
            "spiny chromis",
            "silver-eye",
            "platyfish",
            "california sea lion",
            "This organism does not exist",
        ]
    )

    sp = bt_base.Organism(source="ensembl")
    inspected_df = sp.inspect(df.index, field=sp.name, return_df=True)

    inspect = inspected_df["__validated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert inspect.equals(expected_series)


def test_ensembl_organism_organism():
    for sp in ["bacteria", "plants", "fungi", "metazoa"]:
        df = bt_base.Organism(organism=sp).df()
        assert df.shape[0] > 10


def test_ncbitaxon_organism():
    df = bt_base.Organism(source="ncbitaxon").df()
    assert df.shape[0] > 10



import pandas as pd
from lamindb.models import Record

import bionty.base as bt_base

from ._organism import create_or_get_organism_record


def get_source_record(
    registry: type[Record],
    organism: str | Record | None = None,
    source: Record | None = None,
) -> Record:
    """Get a Source record for a given BioRecord model."""
    from .models import Source

    if source is not None:
        return source

    organism_record = create_or_get_organism_record(organism, registry)

    entity_name = registry.__get_name_with_module__()
    filter_kwargs = {"entity": entity_name}
    if isinstance(organism_record, Record):
        filter_kwargs["organism"] = organism_record.name
    elif isinstance(organism, str):
        filter_kwargs["organism"] = organism

    sources = Source.filter(**filter_kwargs).all()
    if len(sources) == 0:
        raise ValueError(f"No source record found for filter {filter_kwargs}")
    if len(sources) == 1:
        return sources.one()

    current_sources = sources.filter(currently_used=True).all()
    if len(current_sources) == 1:
        return current_sources.first()
    elif len(current_sources) > 1:
        if organism is None:
            # for Organism, in most cases we load from the vertebrates source because ncbitaxon is too big
            if entity_name == "bionty.Organism":
                current_sources_vertebrates = current_sources.filter(
                    organism="vertebrates"
                ).all()
                if len(current_sources_vertebrates) > 0:
                    return current_sources_vertebrates.first()
            # return source with organism="all"
            current_sources_all = current_sources.filter(organism="all").all()
            if len(current_sources_all) > 0:
                return current_sources_all.first()
        return current_sources.first()
    else:  # len(current_sources) == 0
        sources_all = sources.filter(organism="all").all()
        if len(sources_all) > 0:
            # return source with organism="all"
            return sources_all.first()
        return sources.first()


def filter_public_df_columns(
    model: type[Record], public_ontology: bt_base.PublicOntology
) -> pd.DataFrame:
    """Filter columns of public ontology to match the model fields."""

    def _prepare_public_df(model: type[Record], bionty_df: pd.DataFrame):
        """Prepare the bionty DataFrame to match the model fields."""
        if bionty_df.empty:
            return bionty_df
        if model.__get_name_with_module__() == "bionty.Gene":
            # groupby ensembl_gene_id and concat ncbi_gene_ids
            groupby_id_col = (
                "ensembl_gene_id" if "ensembl_gene_id" in bionty_df else "stable_id"
            )
            if groupby_id_col not in bionty_df:
                raise ValueError(
                    "public df must contain column 'ensembl_gene_id' or 'stable_id'"
                )
            bionty_df.drop(
                columns=["hgnc_id", "mgi_id", "index"], errors="ignore", inplace=True
            )
            agg_kwags = {}
            if "ncbi_gene_id" in bionty_df:
                bionty_df.drop_duplicates(
                    [groupby_id_col, "ncbi_gene_id"], inplace=True
                )
                bionty_df["ncbi_gene_id"] = bionty_df["ncbi_gene_id"].fillna("")
                bionty_df.rename(
                    columns={"ncbi_gene_id": "ncbi_gene_ids"}, inplace=True
                )
                agg_kwags["ncbi_gene_ids"] = "|".join
            for col in ["symbol", "biotype", "description", "synonyms"]:
                if col in bionty_df:
                    agg_kwags[col] = "first"  # type: ignore
            bionty_df = bionty_df.groupby(groupby_id_col).agg(agg_kwags).reset_index()

        # rename definition to description for the bionty registry in db
        if "definition" in bionty_df:
            bionty_df.rename(columns={"definition": "description"}, inplace=True)
        return bionty_df

    bionty_df = pd.DataFrame()
    if public_ontology is not None:
        model_field_names = {i.name for i in model._meta.fields}
        # parents needs to be added here as relationships aren't in fields
        model_field_names.add("parents")
        bionty_df = _prepare_public_df(model, public_ontology.df().reset_index())
        bionty_df = bionty_df.loc[:, bionty_df.columns.isin(model_field_names)]
    return bionty_df



import pandas as pd
from lamin_utils import logger

from bionty.base.dev._handle_sources import LAMINDB_INSTANCE_LOADED

from ._settings import settings
from .dev._handle_sources import parse_currently_used_sources


def display_available_sources() -> pd.DataFrame:
    """Displays all available sources.

    Example::

        import bionty.base as bt_base

        bt.display_available_sources()
    """
    from .dev._handle_sources import parse_sources_yaml

    return parse_sources_yaml(settings.public_sources).set_index("entity")  # type: ignore


# This function naming is consistent with the `currently_used` field in Source SQL table
# Do not rename!
def display_currently_used_sources(mute: bool = False) -> pd.DataFrame:
    """Displays all currently used sources.

    Active version is unique for entity + organism.

    Example::

        import bionty.base as bt_base

        bt.display_currently_used_sources()
    """
    if LAMINDB_INSTANCE_LOADED():
        if not mute:
            logger.error(
                "You have a LaminDB instance loaded, please run the following to check default sources:\n"
                "    → bt.Source.filter(currently_used=True).df()"
            )

    versions = parse_currently_used_sources(settings.public_sources)

    df_rows = []
    for bionty_class, bionty_class_data in versions.items():
        for organism, organism_data in bionty_class_data.items():
            for source, version in organism_data.items():
                df_rows.append(
                    {
                        "entity": bionty_class,
                        "organism": organism,
                        "name": source,
                        "version": version,
                    }
                )

    return pd.DataFrame(df_rows).set_index("entity")



# Generated by Django 5.2 on 2024-08-09 10:13

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bionty", "0035_alter_protein_gene_symbol"),
    ]

    operations = [
        migrations.AlterField(
            model_name="source",
            name="artifacts",
            field=models.ManyToManyField(
                related_name="_source_artifact_of", to="lamindb.artifact"
            ),
        ),
        migrations.AlterField(
            model_name="source",
            name="dataframe_artifact",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="_source_dataframe_of",
                to="lamindb.artifact",
            ),
        ),
    ]



from __future__ import annotations

from typing import Literal

import pandas as pd

from bionty.base._public_ontology import PublicOntology
from bionty.base.dev._doc_util import _doc_params
from bionty.base.dev._io import s3_bionty_assets
from bionty.base.entities._shared_docstrings import organism_removed


@_doc_params(doc_entities=organism_removed)
class Organism(PublicOntology):
    """Organism.

    1. Organism ontology
    https://www.ensembl.org/index.html

    Args:
        {doc_entities}
    """

    def __init__(
        self,
        taxa: (
            Literal["vertebrates", "bacteria", "fungi", "metazoa", "plants", "all"]
            | None
        ) = None,
        source: Literal["ensembl", "ncbitaxon"] | None = None,
        version: (
            Literal[
                "2023-06-20",
                "release-57",
                "release-108",
                "release-109",
                "release-110",
                "release-111",
                "release-112",
            ]
            | None
        ) = None,
        **kwargs,
    ):
        # To support the organism kwarg being passed in getattr access in other parts of the code
        # https://github.com/laminlabs/bionty/issues/163
        if "organism" in kwargs and taxa is None:
            taxa = kwargs.pop("organism")
        super().__init__(organism=taxa, source=source, version=version, **kwargs)

    def _get_url(self):
        super()._get_url()
        if (
            not self._url
            and self._source == "ensembl"
            and self._version.startswith("release-")
        ):
            self._url = f"https://ftp.ensembl.org/pub/{self._version}/species_EnsemblVertebrates.txt"

    def _load_df(self) -> pd.DataFrame:
        if self.source == "ensembl":
            if not self._local_parquet_path.exists():
                # try to download from s3
                s3_bionty_assets(
                    filename=self._parquet_filename,
                    assets_base_url="s3://bionty-assets",
                    localpath=self._local_parquet_path,
                )

            # try to download from original url
            if not self._local_parquet_path.exists():
                self._url_download(self._url, self._local_ontology_path)  # type:ignore
                df = pd.read_csv(
                    self._local_ontology_path,
                    sep="\t",
                    index_col=False,  # type:ignore
                )
                df.rename(
                    columns={
                        "#name": "name",
                        "species": "scientific_name",
                        "taxonomy_id": "ontology_id",
                    },
                    inplace=True,
                )
                df["name"] = df["name"].str.lower()
                df["ontology_id"] = "NCBITaxon:" + df["ontology_id"].astype(str)
                df["scientific_name"] = df["scientific_name"].apply(
                    lambda x: " ".join(
                        [x.split("_")[0].capitalize()] + x.split("_")[1:]
                    )
                )
                df.to_parquet(self._local_parquet_path)
                return df
            else:
                df = pd.read_parquet(self._local_parquet_path)
                return _standardize_scientific_name(df)
        else:
            return super()._load_df()

    def df(self) -> pd.DataFrame:
        """Pandas DataFrame of the ontology.

        Returns:
            A Pandas DataFrame of the ontology.

        Example::

            import bionty.base as bionty_base

            bt.Organism().df()
        """
        return self._df.set_index("name")


def _standardize_scientific_name(df: pd.DataFrame) -> pd.DataFrame:
    """Standardize scientific name following NCBITaxon convention.

    homo_sapiens -> Homo sapiens
    """
    df["scientific_name"] = df["scientific_name"].apply(
        lambda x: " ".join([x.split("_")[0].capitalize()] + x.split("_")[1:])
    )
    return df



# Generated by Django 5.2 on 2025-01-13 11:55

import lamindb.base.fields
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("bionty", "0045_rename_aux_cellline__aux_rename_aux_cellmarker__aux_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="cellline",
            name="_aux",
            field=lamindb.base.fields.JSONField(
                blank=True, db_default=None, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="cellmarker",
            name="_aux",
            field=lamindb.base.fields.JSONField(
                blank=True, db_default=None, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="celltype",
            name="_aux",
            field=lamindb.base.fields.JSONField(
                blank=True, db_default=None, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="developmentalstage",
            name="_aux",
            field=lamindb.base.fields.JSONField(
                blank=True, db_default=None, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="disease",
            name="_aux",
            field=lamindb.base.fields.JSONField(
                blank=True, db_default=None, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="ethnicity",
            name="_aux",
            field=lamindb.base.fields.JSONField(
                blank=True, db_default=None, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="experimentalfactor",
            name="_aux",
            field=lamindb.base.fields.JSONField(
                blank=True, db_default=None, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="gene",
            name="_aux",
            field=lamindb.base.fields.JSONField(
                blank=True, db_default=None, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="organism",
            name="_aux",
            field=lamindb.base.fields.JSONField(
                blank=True, db_default=None, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="pathway",
            name="_aux",
            field=lamindb.base.fields.JSONField(
                blank=True, db_default=None, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="phenotype",
            name="_aux",
            field=lamindb.base.fields.JSONField(
                blank=True, db_default=None, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="protein",
            name="_aux",
            field=lamindb.base.fields.JSONField(
                blank=True, db_default=None, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="source",
            name="_aux",
            field=lamindb.base.fields.JSONField(
                blank=True, db_default=None, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="tissue",
            name="_aux",
            field=lamindb.base.fields.JSONField(
                blank=True, db_default=None, default=None, null=True
            ),
        ),
    ]



from __future__ import annotations

from typing import Literal

from bionty.base._public_ontology import PublicOntology
from bionty.base.dev._doc_util import _doc_params

from ._shared_docstrings import doc_entites


@_doc_params(doc_entities=doc_entites)
class Phenotype(PublicOntology):
    """Phenotype.

    1. Human Phenotype Ontology
    https://hpo.jax.org/app/

    2. Mammalian Phenotype Ontology
    https://github.com/mgijax/mammalian-phenotype-ontology

    3. Zebrafish Phenotype Ontology
    https://github.com/obophenotype/zebrafish-phenotype-ontology

    4.Phecodes ICD10 map
    https://phewascatalog.org/phecodes_icd10

    Args:
        {doc_entities}
    """

    def __init__(
        self,
        organism: Literal["human", "mammalian", "zebrafish", "all"] | None = None,
        source: Literal["hp", "phe", "mp", "zp", "pato"] | None = None,
        version: Literal[
            "2023-05-18",
            "2024-03-28",
            "1.2",
            "2022-12-17",
            "2024-01-22",
            "2024-04-18",
            "2023-05-31",
            "2024-02-07",
            "2024-06-18",
            "2023-01-27",
            "2023-04-05",
            "2023-06-17",
            "2024-03-06",
            "2024-04-26",
        ]
        | None = None,
        **kwargs,
    ) -> None:
        super().__init__(
            source=source,
            version=version,
            organism=organism,
            include_id_prefixes={
                "hp": ["HP"],
                "mp": ["MP"],  # mp might require an exclusion prefix for mpath
                "zp": ["ZP"],
                "pato": ["PATO"],
            },
            **kwargs,
        )



import tempfile
from pathlib import Path

import pytest
from bionty.base.dev._io import url_download


@pytest.fixture
def local(tmp_path):
    url = "https://bionty-assets.s3.amazonaws.com/bfxpipelines.json"
    localpath = tmp_path / Path(url).name
    yield localpath, url
    if localpath.exists():
        localpath.unlink()


def test_url_download(local):
    localpath = local[0]
    url = local[1]
    assert not localpath.exists()

    downloaded_path = Path(url_download(url=url, localpath=localpath))
    assert downloaded_path.exists()


def test_local_file():
    with tempfile.TemporaryDirectory() as temp_dir:
        local_file = Path(temp_dir) / "test.txt"
        target_file = Path(temp_dir) / "downloaded.txt"
        test_content = "temporary file"

        local_file.write_text(test_content)
        assert local_file.exists(), "Test file was not created"

        downloaded_path = Path(
            url_download(url=f"file://{local_file}", localpath=target_file)
        )

        assert downloaded_path.exists(), "Downloaded file not found"
        assert downloaded_path.read_text() == test_content, "Content mismatch"

        if downloaded_path.exists():
            downloaded_path.unlink()



from __future__ import annotations

from lamin_utils import logger

from bionty.models import Organism


class Settings:
    """Settings.

    Directly use `bt.settings` rather than instantiating this class yourself.
    """

    def __init__(self):
        self._organism = "human"

    @property
    def organism(self) -> Organism | None:
        """Default organism argument (default `"human"`).

        Default organism to use in cases of ambiguity. For instance, gene symbols are duplicated across organisms and need to be disambiguated.

        Examples:

            ::

                bionty.settings.organism = "mouse"
        """
        if isinstance(self._organism, str):
            self.organism = self._organism  # type: ignore
        return self._organism

    @organism.setter
    def organism(self, name: str | Organism):
        if isinstance(name, Organism):
            self._organism = name
        else:
            import lamindb as ln

            # do not show the validated message for organism
            verbosity = ln.settings.verbosity
            ln.settings.verbosity = 1
            organisms = Organism.from_values([name])
            ln.settings.verbosity = verbosity
            if len(organisms) == 0:
                raise ValueError(
                    f"No organism with name='{name}' is found, please create a organism record!"
                )
            else:
                organism = organisms[0]
            if organism._state.adding:  # type:ignore
                organism.save()  # type:ignore
            self._organism = organism


settings = Settings()
settings.__doc__ = """Global :class:`~bionty.core.Settings`."""



import bionty.base as bt_base
import pandas as pd


def test_cl_celltype_inspect_name():
    df = pd.DataFrame(
        index=[
            "Boettcher cell",
            "bone marrow cell",
            "interstitial cell of ovary",
            "pancreatic ductal cell",
            "This cell type does not exist",
        ]
    )

    ct = bt_base.CellType(source="cl")
    inspected_df = ct.inspect(df.index, field=ct.name, return_df=True)

    inspect = inspected_df["__validated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert inspect.equals(expected_series)



from __future__ import annotations

from typing import TYPE_CHECKING

from lamin_utils import logger

from bionty._organism import create_or_get_organism_record, is_organism_required

if TYPE_CHECKING:
    from collections.abc import Iterable

    import pandas as pd
    from lamindb.models import Record

    from bionty.models import BioRecord, Organism, Source


def get_all_ancestors(df: pd.DataFrame, ontology_ids: Iterable[str]) -> set[str]:
    ancestors = set()
    stack = list(ontology_ids)
    while stack:
        onto_id = stack.pop()
        try:
            parents = df.at[onto_id, "parents"]
            for parent in parents:
                if parent not in ancestors:
                    ancestors.add(parent)
                    stack.append(parent)
        except KeyError:
            logger.warning(f"ontology ID {onto_id} not found in DataFrame")
    return ancestors


def prepare_dataframe(df: pd.DataFrame) -> pd.DataFrame:
    df = df.reset_index()
    if "ontology_id" in df.columns:
        return df.set_index("ontology_id")
    return df


def get_new_ontology_ids(
    registry: type[BioRecord], ontology_ids: Iterable[str], df: pd.DataFrame
) -> tuple[set[str], set[str]]:
    all_ontology_ids = set(ontology_ids) | get_all_ancestors(df, ontology_ids)
    existing_ontology_ids = set(
        registry.filter(ontology_id__in=all_ontology_ids).values_list(
            "ontology_id", flat=True
        )
    )
    return (all_ontology_ids - existing_ontology_ids), all_ontology_ids


def create_records(
    registry: type[BioRecord],
    df: pd.DataFrame,
    source_record: Source,
    organism: str | Organism | None = None,
) -> list[Record]:
    """Bulk-create records from a DataFrame skipping validation."""
    df = df.reset_index()
    df = df.rename(columns={"definition": "description"})

    if "parents" in df.columns:
        df = df.drop(columns=["parents"])

    df_records = df.to_dict(orient="records")

    if organism is None and is_organism_required(registry):
        organism = source_record.organism

    organism = create_or_get_organism_record(organism=organism, registry=registry)

    valid_fields = [f.name for f in registry._meta.fields]
    records = [
        registry(
            **{k: v for k, v in record.items() if k in valid_fields},
            **({"organism": organism} if "organism" in valid_fields else {}),
            source_id=source_record.id,
            _skip_validation=True,
        )
        for record in df_records
    ]

    return records


def create_link_records(
    registry: type[BioRecord], df: pd.DataFrame, records: list[Record]
) -> list[Record]:
    """Create link records.

    Args:
        registry: The model class of the records.
        df: The DataFrame with all ontology IDs and their parents.
        records: All records of the ontology.
    """
    source = records[0].source
    linkorm = registry.parents.through
    link_records = []
    registry_name_lower = registry.__name__.lower()

    # Create a dictionary for quick lookups
    record_dict = {r.ontology_id: r for r in records if r.source_id == source.id}

    for child_id, parents_ids in df["parents"].items():
        if len(parents_ids) == 0:
            continue
        child_record = record_dict.get(child_id)
        if not child_record:
            continue
        for parent_id in parents_ids:
            parent_record = record_dict.get(parent_id)
            if parent_record:
                link_records.append(
                    linkorm(
                        **{
                            f"from_{registry_name_lower}": child_record,
                            f"to_{registry_name_lower}": parent_record,
                        }
                    )
                )
    return link_records


def check_source_in_db(
    registry: type[BioRecord],
    source: Source,
    n_all: int = None,
    n_in_db: int = None,
) -> None:
    if not hasattr(registry, "source_id"):
        logger.warning(f"no `source` field in the registry {registry.__name__}!")
    else:
        n_all = n_all or registry.public(source=source).df().shape[0]
        # all records of the source in the database
        n_in_db = n_in_db or registry.filter(source=source).count()
        if n_in_db >= n_all:
            # make sure in_db is set to True if all records are in the database
            source.in_db = True
            source.save()
        else:
            source.in_db = False
            source.save()


def add_ontology_from_df(
    registry: type[BioRecord],
    ontology_ids: list[str] | None = None,
    organism: str | Organism | None = None,
    source: Source | None = None,
    ignore_conflicts: bool = True,
) -> None:
    """Add ontology records from source to the database based on ontology ids."""
    import lamindb as ln

    from bionty._source import get_source_record

    source_record = get_source_record(registry, organism=organism, source=source)
    public = registry.public(source=source_record)
    df = prepare_dataframe(public.df())

    if ontology_ids is None:
        logger.info(
            f"importing {registry.__name__} records from {public.source}, {public.version}"
        )
        df_new = df_all = df
    else:
        new_ontology_ids, all_ontology_ids = get_new_ontology_ids(
            registry, ontology_ids, df
        )
        df_new = df[df.index.isin(new_ontology_ids)]
        df_all = df[df.index.isin(all_ontology_ids)]

    # do not create records from obsolete terms
    if hasattr(registry, "ontology_id"):
        df_all = df_all[~df_all["name"].fillna("").str.startswith("obsolete")]

    n_all = df_all.shape[0]
    if n_all == 0:
        raise ValueError("No valid records to add!")

    # all records of the source in the database
    all_records = registry.filter(source=source_record).all()
    n_in_db = all_records.count()

    check_source_in_db(
        registry=registry,
        source=source_record,
        n_all=n_all,
        n_in_db=n_in_db,
    )

    records = create_records(registry, df_new, source_record, organism)
    new_records = [r for r in records if r._state.adding]
    if ontology_ids is None:
        logger.info(f"adding {len(new_records)} new records")
    registry.objects.bulk_create(new_records, ignore_conflicts=ignore_conflicts)

    all_records = registry.filter(
        source=source_record
    ).all()  # need to update all_records after bulk_create
    if hasattr(registry, "parents"):
        link_records = create_link_records(registry, df_all, all_records)
        new_link_records = [r for r in link_records if r._state.adding]
        if ontology_ids is None:
            logger.info(f"adding {len(new_link_records)} parents/children links")
        ln.save(new_link_records, ignore_conflicts=ignore_conflicts)

    if ontology_ids is None:
        logger.success("import is completed!")
        source_record.in_db = True
        source_record.save()


# used in save() to bulk save parents
def add_ontology(
    records: list[BioRecord],
    organism: str | Organism | None = None,
    source: Source | None = None,
    ignore_conflicts: bool = True,
) -> None:
    """Add ontology records from source to the database based on ontology ids."""
    registry = records[0]._meta.model
    source = source or records[0].source
    if (
        hasattr(registry, "organism_id")
        and not registry._meta.get_field("organism_id").null
    ):
        organism = organism or records[0].organism
    ontology_ids = [r.ontology_id for r in records]
    add_ontology_from_df(
        registry=registry,
        ontology_ids=ontology_ids,
        organism=organism,
        source=source,
        ignore_conflicts=ignore_conflicts,
    )



import os

import nox
from laminci.nox import build_docs, install_lamindb, run, run_pre_commit

nox.options.default_venv_backend = "none"

IS_PR = os.getenv("GITHUB_EVENT_NAME") != "push"


@nox.session
def lint(session: nox.Session) -> None:
    run_pre_commit(session)


@nox.session
@nox.parametrize("group", ["bionty-base", "bionty-core", "bionty-docs"])
def build(session: nox.Session, group: str):
    branch = "main" if IS_PR else "release"  # point back to "release"
    install_lamindb(session, branch=branch)
    run(session, "uv pip install --system wetlab")
    session.run(*"uv pip install --system -e .[dev]".split())

    coverage_args = "--cov=bionty --cov-append --cov-report=term-missing"
    if group == "bionty-base":
        session.run(*f"pytest {coverage_args} ./tests/base".split())
    elif group == "bionty-core":
        session.run(*f"pytest {coverage_args} ./tests/core".split())
    elif group == "bionty-docs":
        session.run(*f"pytest -s {coverage_args} ./docs/guide".split())
        run(session, "lamin init --storage ./docsbuild --modules bionty")
        build_docs(session, strict=False)



# Generated by Django 5.2 on 2025-01-07 14:37

import django.db.models.deletion
import lamindb.base.fields
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bionty", "0041_alter_artifactcellline_artifact_and_more"),
        ("lamindb", "0071_lamindbv1_migrate_schema"),
    ]

    operations = [
        migrations.AddField(
            model_name="cellline",
            name="_branch_code",
            field=models.SmallIntegerField(db_index=True, default=1, db_default=1),
        ),
        migrations.AddField(
            model_name="cellmarker",
            name="_branch_code",
            field=models.SmallIntegerField(db_index=True, default=1, db_default=1),
        ),
        migrations.AddField(
            model_name="celltype",
            name="_branch_code",
            field=models.SmallIntegerField(db_index=True, default=1, db_default=1),
        ),
        migrations.AddField(
            model_name="developmentalstage",
            name="_branch_code",
            field=models.SmallIntegerField(db_index=True, default=1, db_default=1),
        ),
        migrations.AddField(
            model_name="disease",
            name="_branch_code",
            field=models.SmallIntegerField(db_index=True, default=1, db_default=1),
        ),
        migrations.AddField(
            model_name="ethnicity",
            name="_branch_code",
            field=models.SmallIntegerField(db_index=True, default=1, db_default=1),
        ),
        migrations.AddField(
            model_name="experimentalfactor",
            name="_branch_code",
            field=models.SmallIntegerField(db_index=True, default=1, db_default=1),
        ),
        migrations.AddField(
            model_name="gene",
            name="_branch_code",
            field=models.SmallIntegerField(db_index=True, default=1, db_default=1),
        ),
        migrations.AddField(
            model_name="organism",
            name="_branch_code",
            field=models.SmallIntegerField(db_index=True, default=1, db_default=1),
        ),
        migrations.AddField(
            model_name="pathway",
            name="_branch_code",
            field=models.SmallIntegerField(db_index=True, default=1, db_default=1),
        ),
        migrations.AddField(
            model_name="phenotype",
            name="_branch_code",
            field=models.SmallIntegerField(db_index=True, default=1, db_default=1),
        ),
        migrations.AddField(
            model_name="protein",
            name="_branch_code",
            field=models.SmallIntegerField(db_index=True, default=1, db_default=1),
        ),
        migrations.AddField(
            model_name="source",
            name="_branch_code",
            field=models.SmallIntegerField(db_index=True, default=1, db_default=1),
        ),
        migrations.AddField(
            model_name="tissue",
            name="_branch_code",
            field=models.SmallIntegerField(db_index=True, default=1, db_default=1),
        ),
        migrations.AddField(
            model_name="cellline",
            name="aux",
            field=models.JSONField(default=None, db_default=None, null=True),
        ),
        migrations.AddField(
            model_name="cellmarker",
            name="aux",
            field=models.JSONField(default=None, db_default=None, null=True),
        ),
        migrations.AddField(
            model_name="celltype",
            name="aux",
            field=models.JSONField(default=None, db_default=None, null=True),
        ),
        migrations.AddField(
            model_name="developmentalstage",
            name="aux",
            field=models.JSONField(default=None, db_default=None, null=True),
        ),
        migrations.AddField(
            model_name="disease",
            name="aux",
            field=models.JSONField(default=None, db_default=None, null=True),
        ),
        migrations.AddField(
            model_name="ethnicity",
            name="aux",
            field=models.JSONField(default=None, db_default=None, null=True),
        ),
        migrations.AddField(
            model_name="experimentalfactor",
            name="aux",
            field=models.JSONField(default=None, db_default=None, null=True),
        ),
        migrations.AddField(
            model_name="gene",
            name="aux",
            field=models.JSONField(default=None, db_default=None, null=True),
        ),
        migrations.AddField(
            model_name="organism",
            name="aux",
            field=models.JSONField(default=None, db_default=None, null=True),
        ),
        migrations.AddField(
            model_name="pathway",
            name="aux",
            field=models.JSONField(default=None, db_default=None, null=True),
        ),
        migrations.AddField(
            model_name="phenotype",
            name="aux",
            field=models.JSONField(default=None, db_default=None, null=True),
        ),
        migrations.AddField(
            model_name="protein",
            name="aux",
            field=models.JSONField(default=None, db_default=None, null=True),
        ),
        migrations.AddField(
            model_name="source",
            name="aux",
            field=models.JSONField(default=None, db_default=None, null=True),
        ),
        migrations.AddField(
            model_name="tissue",
            name="aux",
            field=models.JSONField(default=None, db_default=None, null=True),
        ),
        migrations.RemoveField(
            model_name="cellline",
            name="_previous_runs",
        ),
        migrations.RemoveField(
            model_name="cellmarker",
            name="_previous_runs",
        ),
        migrations.RemoveField(
            model_name="celltype",
            name="_previous_runs",
        ),
        migrations.RemoveField(
            model_name="developmentalstage",
            name="_previous_runs",
        ),
        migrations.RemoveField(
            model_name="disease",
            name="_previous_runs",
        ),
        migrations.RemoveField(
            model_name="ethnicity",
            name="_previous_runs",
        ),
        migrations.RemoveField(
            model_name="experimentalfactor",
            name="_previous_runs",
        ),
        migrations.RemoveField(
            model_name="gene",
            name="_previous_runs",
        ),
        migrations.RemoveField(
            model_name="organism",
            name="_previous_runs",
        ),
        migrations.RemoveField(
            model_name="pathway",
            name="_previous_runs",
        ),
        migrations.RemoveField(
            model_name="phenotype",
            name="_previous_runs",
        ),
        migrations.RemoveField(
            model_name="protein",
            name="_previous_runs",
        ),
        migrations.RemoveField(
            model_name="source",
            name="_previous_runs",
        ),
        migrations.RemoveField(
            model_name="tissue",
            name="_previous_runs",
        ),
    ]



import tempfile
from pathlib import Path

import pytest
from bionty.base.dev._handle_sources import (
    parse_currently_used_sources,
    parse_sources_yaml,
)


@pytest.fixture(scope="function")
def versions_yaml_replica():
    input_file_content = """
    version: "0.2.0"
    Organism:
      ensembl:
        all:
          release-108:
            url: https://ftp.ensembl.org/pub/release-108/species_EnsemblVertebrates.txt
        name: Ensembl
        website: https://www.ensembl.org/index.html
    Gene:
      ensembl:
        human:
          release-108:
            url: https://ftp.ensembl.org/pub/release-108/mysql/homo_sapiens_core_108_38/
          release-107:
            url: https://ftp.ensembl.org/pub/release-107/mysql/homo_sapiens_core_107_38/
        mouse:
          release-108:
            url: https://ftp.ensembl.org/pub/release-108/mysql/mus_musculus_core_108_39/
        name: Ensembl
        website: https://www.ensembl.org/index.html
    CellType:
      cl:
        name: Cell Ontology
        website: https://obophenotype.github.io/cell-ontology/
        all:
          2023-02-15:
            url: http://purl.obolibrary.org/obo/cl/releases/2023-02-15/cl-base.owl
          2022-08-16:
            url: http://purl.obolibrary.org/obo/cl/releases/2022-08-16/cl.owl
    """
    with tempfile.NamedTemporaryFile(mode="w+", delete=False) as f:
        f.write(input_file_content)
        f.flush()
        yield f.name

    Path(f.name).unlink()


def test_parse_versions_yaml(versions_yaml_replica):
    parsed_df = parse_sources_yaml(versions_yaml_replica)
    assert parsed_df.shape == (6, 7)
    assert all(
        parsed_df["entity"].values
        == ["Organism", "Gene", "Gene", "Gene", "CellType", "CellType"]
    )
    assert all(
        parsed_df["organism"].values == ["all", "human", "human", "mouse", "all", "all"]
    )
    assert all(
        parsed_df["name"].values
        == ["ensembl", "ensembl", "ensembl", "ensembl", "cl", "cl"]
    )


def test_parse_current_versions(versions_yaml_replica):
    expected = {
        "Organism": {"all": {"ensembl": "release-108"}},
        "Gene": {
            "human": {"ensembl": "release-108"},
            "mouse": {"ensembl": "release-108"},
        },
        "CellType": {"all": {"cl": "2023-02-15"}},
    }

    assert parse_currently_used_sources(versions_yaml_replica) == expected



from __future__ import annotations

from typing import Literal

from bionty.base._public_ontology import PublicOntology
from bionty.base.dev._doc_util import _doc_params

from ._shared_docstrings import doc_entites


@_doc_params(doc_entities=doc_entites)
class CellLine(PublicOntology):
    """Cell line.

    1. Cell Line Ontology
    https://github.com/CLO-ontology/CLO

    2. DepMap
    https://depmap.org

    Args:
        {doc_entities}
    """

    def __init__(
        self,
        organism: Literal["all"] | None = None,
        source: Literal["clo", "depmap"] | None = None,
        version: Literal["2022-03-21", "2024-Q2"] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(
            source=source,
            version=version,
            organism=organism,
            include_id_prefixes={"clo": ["CLO"]},
            **kwargs,
        )



import bionty.base as bt_base
import pandas as pd


def test_pw_go_inspect_ontology_id():
    df = pd.DataFrame(
        index=[
            "GO:1905210",
            "GO:1905211",
            "GO:1905212",
            "GO:1905208",
            "This pathway does not exist",
        ]
    )

    pw = bt_base.Pathway(source="go")
    inspected_df = pw.inspect(df.index, pw.ontology_id, return_df=True)

    inspect = inspected_df["__validated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert inspect.equals(expected_series)


def test_pw_pathway_inspect_name():
    df = pd.DataFrame(
        index=[
            "Toll-like receptor 9 signaling pathway",
            "Toll-like receptor TLR1:TLR2 signaling pathway",
            "classic metabolic pathway",
            "regulatory pathway",
            "This pathway does not exist",
        ]
    )

    pw = bt_base.Pathway(source="pw")
    inspected_df = pw.inspect(df.index, field=pw.name, return_df=True)

    inspect = inspected_df["__validated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert inspect.equals(expected_series)



import bionty.base as bt_base
import pandas as pd


def test_efo_experimental_factor_inspect_ontology_id():
    df = pd.DataFrame(
        index=[
            "EFO:1002048",
            "EFO:1002050",
            "EFO:1002047",
            "EFO:1002049",
            "This readout does not exist",
        ]
    )

    ro = bt_base.ExperimentalFactor(source="efo")
    inspected_df = ro.inspect(df.index, ro.ontology_id, return_df=True)

    inspect = inspected_df["__validated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert inspect.equals(expected_series)



from __future__ import annotations

from typing import Literal

from bionty.base._public_ontology import PublicOntology
from bionty.base.dev._doc_util import _doc_params

from ._shared_docstrings import doc_entites


@_doc_params(doc_entities=doc_entites)
class Disease(PublicOntology):
    """Disease ontologies.

    1. Mondo
    Edits of terms are coordinated and reviewed on:
    https://github.com/monarch-initiative/mondo

    2. Human Disease Ontology
    Edits of terms are coordinated and reviewed on:
    https://github.com/DiseaseOntology/HumanDiseaseOntology

    3. International Classification of Diseases (ICD)
    Edits of terms are coordinated and reviewed on:
    https://www.who.int/standards/classifications/classification-of-diseases

    Args:
        {doc_entities}
    """

    def __init__(
        self,
        organism: Literal["all", "human"] | None = None,
        source: Literal["mondo", "doid", "icd"] | None = None,
        version: Literal[
            "icd-9-2011",
            "icd-10-2020",
            "icd-10-2024",
            "icd-11-2023",
            "2023-01-30",
            "2023-03-31",
            "2024-01-31",
            "2024-05-29",
            "2022-10-11",
            "2023-02-06",
            "2023-04-04",
            "2023-08-02",
            "2024-01-03",
            "2024-02-06",
            "2024-05-08",
            "2024-06-04",
            "2024-08-06",
        ]
        | None = None,
        **kwargs,
    ) -> None:
        super().__init__(
            source=source,
            version=version,
            organism=organism,
            include_id_prefixes={"mondo": ["MONDO"]},
            **kwargs,
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from . import ids

if TYPE_CHECKING:
    from types import ModuleType

    from .models import BioRecord


def encode_uid(registry: type[BioRecord], kwargs: dict):
    from lamindb.models import Record

    if kwargs.get("uid") is not None:
        # if uid is passed, no encoding is needed
        return kwargs
    name = registry.__name__.lower()
    if hasattr(registry, "organism_id"):
        organism = kwargs.get("organism")
        if organism is None:
            if kwargs.get("organism_id") is not None:
                from .models import Organism

                organism = Organism.get(kwargs.get("organism_id")).name
        elif isinstance(organism, Record):
            organism = organism.name
    else:
        organism = ""

    if hasattr(registry, "_ontology_id_field"):
        ontology_id_field = registry._ontology_id_field
    else:
        ontology_id_field = "ontology_id"
    if hasattr(registry, "_name_field"):
        name_field = registry._name_field
    else:
        name_field = "name"

    str_to_encode = None
    if name == "source":
        str_to_encode = f"{kwargs.get('entity', '')}{kwargs.get('name', '')}{kwargs.get('organism', '')}{kwargs.get('version', '')}"
    elif name == "gene":  # gene has multiple id fields
        str_to_encode = kwargs.get(ontology_id_field)
        if str_to_encode is None or str_to_encode == "":
            str_to_encode = kwargs.get("stable_id")
        if str_to_encode is None or str_to_encode == "":
            str_to_encode = f"{kwargs.get(name_field)}{organism}"  # name + organism
        if str_to_encode is None or str_to_encode == "":
            raise AssertionError(
                f"must provide {ontology_id_field}, stable_id or {name_field}"
            )
    else:
        str_to_encode = kwargs.get(ontology_id_field)
        if str_to_encode is None or str_to_encode == "":
            str_to_encode = f"{kwargs.get(name_field)}{organism}"  # name + organism
        if str_to_encode is None or str_to_encode == "":
            raise AssertionError(f"must provide {ontology_id_field} or {name_field}")

    if str_to_encode is not None and len(str_to_encode) > 0:
        try:
            id_encoder = getattr(ids, name)
        except Exception:
            if ontology_id_field == "ontology_id":
                id_encoder = ids.ontology
            else:
                return kwargs
        kwargs["uid"] = id_encoder(str_to_encode)
    return kwargs


def lookup2kwargs(record: BioRecord, *args, **kwargs) -> dict:
    """Pass bionty search/lookup results."""
    from ._organism import create_or_get_organism_record
    from ._source import get_source_record

    arg = args[0]
    if isinstance(arg, tuple):
        bionty_kwargs = arg._asdict()  # type:ignore
    else:
        bionty_kwargs = arg[0]._asdict()

    if len(bionty_kwargs) > 0:
        # add organism and source
        organism_record = create_or_get_organism_record(
            registry=record.__class__, organism=kwargs.get("organism")
        )
        if organism_record is not None:
            bionty_kwargs["organism"] = organism_record
        bionty_kwargs["source"] = get_source_record(
            registry=record.__class__,
            organism=organism_record,
            source=kwargs.get("source"),
        )

        model_field_names = {i.name for i in record._meta.fields}
        model_field_names.add("parents")
        bionty_kwargs = {
            k: v for k, v in bionty_kwargs.items() if k in model_field_names
        }
    return encode_uid(registry=record.__class__, kwargs=bionty_kwargs)


def list_biorecord_models(schema_module: ModuleType):
    """List all BioRecord models in a given schema module."""
    import inspect

    import lamindb as ln  # needed here

    from .models import BioRecord

    return [
        attr
        for attr in dir(schema_module.models)
        if inspect.isclass(getattr(schema_module.models, attr))
        and issubclass(getattr(schema_module.models, attr), BioRecord)
    ]
